@@ -1,4 +1,3 @@
-# item.py
 """ Functions for handling content items """
 
 import email
@@ -51,9 +50,7 @@ class Entry(caching.Memoizable):
 
         assert create_key == Entry.load.__name__, "Entry must be created with Entry.load()"
 
-        LOGGER.debug('init entry %d', record.id)
         self._record = record   # index record
-
         self._fingerprint = model.FileFingerprint.get(file_path=record.file_path)
         LOGGER.debug('loaded entry %d, fingerprint=%s', record.id,
                      self._fingerprint.fingerprint if self._fingerprint else None)
@@ -181,7 +178,7 @@ class Entry(caching.Memoizable):
                 raise ValueError(f"Unknown paging type '{paging}'")
 
             if tag:
-                args['tag'] = tag
+                args['tag'] = list(utils.TagSet(utils.as_list(tag)).keys())
 
             return flask.url_for('category', **args, _external=absolute)
 
@@ -194,7 +191,7 @@ class Entry(caching.Memoizable):
 
     @cached_property
     def tags(self) -> typing.List[str]:
-        """ Get the original (non-normalized) tags for the entry """
+        """ Get the original (non-normalized, non-folded) tags for the entry """
         return self.get_all('Tag')
 
     @cached_property
@@ -367,7 +364,6 @@ class Entry(caching.Memoizable):
         body, more, is_markdown = self._entry_content
 
         def _footnotes(**kwargs) -> str:
-            LOGGER.debug("rendering footnotes; args=%s", kwargs)
             return self._get_footnotes(body, more, kwargs)
 
         if is_markdown:
@@ -376,13 +372,10 @@ class Entry(caching.Memoizable):
 
             if ((body_count and body_count.footnote)
                     or (more_count and more_count.footnote)):
-                LOGGER.debug("We definitely have footnotes")
                 return TrueCallableProxy(_footnotes)
             if body_count is None or more_count is None:
-                LOGGER.debug("We might have footnotes")
                 return CallableProxy(_footnotes)
 
-        LOGGER.debug("There are definitely no footnotes")
         return CallableValue('')
 
     @cached_property
@@ -391,7 +384,6 @@ class Entry(caching.Memoizable):
         body, more, is_markdown = self._entry_content
 
         def _toc(max_depth=None, **kwargs) -> str:
-            LOGGER.debug("rendering table of contents; max_depth=%s kwargs=%s", max_depth, kwargs)
             return self._get_toc(body, more, max_depth, kwargs)
 
         if is_markdown:
@@ -400,14 +392,11 @@ class Entry(caching.Memoizable):
 
             if ((body_count and body_count.toc)
                     or (more_count and more_count.toc)):
-                LOGGER.debug("We definitely have a ToC")
                 return TrueCallableProxy(_toc)
 
             if body_count is None or more_count is None:
-                LOGGER.debug("We might have a ToC")
                 return CallableProxy(_toc)
 
-        LOGGER.debug("There is definitely no TOC")
         return CallableValue('')
 
     @cached_property
@@ -417,8 +406,6 @@ class Entry(caching.Memoizable):
         def _get_card(**kwargs) -> str:
             """ Render out the tags for a Twitter/OpenGraph card for this entry. """
 
-            LOGGER.debug("rendering card; args=%s", kwargs)
-
             def og_tag(key, val) -> str:
                 """ produce an OpenGraph tag with the given key and value """
                 return utils.make_tag('meta', {'property': key, 'content': val}, start_end=True)
@@ -507,11 +494,12 @@ class Entry(caching.Memoizable):
         """
         # pylint:disable=too-many-arguments
         if is_markdown:
-            if 'footnotes_link' not in args:
-                args['footnotes_link'] = self.link(absolute=args.get('absolute'))
-
-            if 'toc_link' not in args:
-                args['toc_link'] = self.link(absolute=args.get('absolute'))
+            # Set defaults for the ID link generators, so permalinks from category
+            # pages work correctly
+            default_link = self.link(absolute=args.get('absolute'))
+            for link_flag in ('footnotes_link', 'toc_link', 'code_number_links'):
+                if link_flag not in args or args[link_flag] is True:
+                    args[link_flag] = default_link
 
             return markdown.to_html(
                 text,
@@ -524,11 +512,16 @@ class Entry(caching.Memoizable):
                 counter=counter
             )
 
-        return html_entry.process(
+        text = html_entry.process(
             text,
             args,
             search_path=self.search_path)
 
+        if not args.get('markup', True):
+            text = html_entry.strip_html(text)
+
+        return text
+
     @cached_property
     def attachments(self) -> typing.Callable[..., typing.List]:
         """ Returns a view of entries that are attached to this one. Takes the
@@ -540,7 +533,10 @@ class Entry(caching.Memoizable):
                                          })
             if order:
                 query = query.order_by(*queries.ORDER_BY[order])
-            return [Entry.load(e) for e in query]
+            cur_user = user.get_active()
+
+            return [Entry.load(e) for e in query
+                    if e.is_authorized(cur_user) or tokens.request(cur_user)]
 
         return CallableProxy(_get_attachments)
 
@@ -610,9 +606,7 @@ class Entry(caching.Memoizable):
             raise ValueError("Unknown content section " + section)
 
         if text:
-            LOGGER.debug("Getting counters for %s,%s", section, footnotes)
             counter = markdown.get_counters(text, args)
-            LOGGER.debug("Caching %s:%s -> %s", section, footnotes, counter)
             self._counters[(section, footnotes)] = counter
             return counter
 
@@ -621,7 +615,6 @@ class Entry(caching.Memoizable):
     def _set_counter(self, section, args, counter: markdown.ItemCounter):
         """ Register the counts that we already know """
         footnotes = 'footnotes' in args.get('markdown_extensions', config.markdown_extensions)
-        LOGGER.debug("Registering %s:%s -> %s", section, footnotes, counter)
         self._counters[(section, footnotes)] = counter
 
     def __getattr__(self, name):
@@ -641,7 +634,6 @@ class Entry(caching.Memoizable):
         category = kwargs.get('category', self._record.category)
         return {
             'category': category,
-            'recurse': kwargs.get('recurse', 'category' in kwargs)
         }
 
     def get(self, name, default=None) -> typing.Optional[str]:
@@ -658,6 +650,28 @@ class Entry(caching.Memoizable):
         # pylint:disable=protected-access
         return isinstance(other, Entry) and (other is self or other._record == self._record)
 
+    @staticmethod
+    def filter_auth(entries, count=None, unauthorized=0):
+        """ Filter a list of entries based on authorization, with a maximum
+        unauthorized entry count """
+
+        result: typing.List[Entry] = []
+        cur_user = user.get_active()
+        for record in entries:
+            if count is not None and len(result) >= count:
+                break
+
+            auth = record.is_authorized(cur_user)
+            if auth or unauthorized:
+                result.append(Entry.load(record))
+                if not auth and unauthorized is not True:
+                    unauthorized -= 1
+
+            if not auth:
+                tokens.request(cur_user)
+
+        return result
+
 
 def get_entry_id(entry, fullpath, assign_id) -> typing.Optional[int]:
     """ Get or generate an entry ID for an entry """
@@ -824,7 +838,7 @@ def scan_file(fullpath: str, relpath: typing.Optional[str], fixup_pass: int) ->
         record = model.Entry(id=entry_id, **values)
 
     # Update the entry ID
-    if str(record.id) != entry['Entry-ID']:
+    if str(record.id) != entry['Entry-ID']:  # pylint:disable=no-member
         del entry['Entry-ID']
         entry['Entry-ID'] = str(record.id)
         fixup_needed = True
@@ -855,34 +869,45 @@ def scan_file(fullpath: str, relpath: typing.Optional[str], fixup_pass: int) ->
 
     with orm.db_session:
         set_tags = {
-            t.casefold(): t
-            for t in entry.get_all('Tag', [])
-            + entry.get_all('Hidden-Tag', [])
+            utils.tag_key(t[0]): t
+            for t in [(k, True) for k in entry.get_all('Hidden-Tag', [])]
+            + [(k, False) for k in entry.get_all('Tag', [])]
         }
-        LOGGER.debug("set_tags %s", set_tags)
         remove_tags = []
 
-        for tag in record.tags:
-            LOGGER.debug("  has tag %s,%s", tag.key, tag.name)
-            if tag.key not in set_tags:
-                remove_tags.append(tag)
+        for etag in record.tags:
+            LOGGER.debug("  has tag %s,%s", etag.tag.key, etag.tag.name)
+            if etag.tag.key not in set_tags:
+                remove_tags.append(etag)
         LOGGER.debug("set_tags %s remove_tags %s", set_tags, remove_tags)
 
-        for tag in remove_tags:
-            record.tags.remove(tag)
+        for etag in remove_tags:
+            tag = etag.tag
+            etag.delete()
             if len(tag.entries) == 0:
                 LOGGER.debug("tag %s/%s entry count went to 0", tag.key, tag.name)
                 tag.delete()
 
-        for (key, name) in set_tags.items():
+        for (key, tag) in set_tags.items():
+            name, hidden = tag
+
+            # get the underlying tag object
             tag_record = model.EntryTag.get(key=key)
             if not tag_record:
                 LOGGER.debug("creating tag %s/%s", key, name)
                 tag_record = model.EntryTag(key=key, name=name)
-            elif name != tag_record.name:
-                LOGGER.debug("updating tag name %s/%s -> %s", key, tag_record.name, name)
+            elif name != tag_record.name and name != key and not hidden:
+                LOGGER.debug("updating tag name %s/%s -> %s",
+                             key, tag_record.name, name)
                 tag_record.name = name
-            record.tags.add(tag_record)
+
+            # get the tag placement object
+            etag = model.EntryTagged.get(tag=tag_record, entry=record)
+            if not etag:
+                etag = model.EntryTagged(tag=tag_record, entry=record, hidden=hidden)
+            else:
+                etag.hidden = hidden
+            record.tags.add(etag)
 
         orm.commit()
 
@@ -931,6 +956,9 @@ def scan_file(fullpath: str, relpath: typing.Optional[str], fixup_pass: int) ->
         LOGGER.info("Fixing up entry %s", fullpath)
         result = save_file(fullpath, entry, check_fingerprint)
 
+    # register with the search index
+    flask.current_app.search_index.update(record, entry)
+
     return result
 
 
