@@ -81,7 +81,7 @@ class HtmlCodeFormatter(pygments.formatters.HtmlFormatter):  # pylint:disable=no
 
     def __init__(self,
                  line_id_prefix: str,
-                 link_base: str,
+                 link_base: typing.Union[str, bool],
                  *args, **kwargs):
         super().__init__(*args, **kwargs)
         self.line_id_prefix = line_id_prefix
@@ -99,27 +99,30 @@ class HtmlCodeFormatter(pygments.formatters.HtmlFormatter):  # pylint:disable=no
                 line_number += 1
                 line_id = "{}L{}".format(self.line_id_prefix, line_number)
 
-                yield 1, (utils.make_tag('span', {
-                    'class': 'line',
-                    'id': line_id
-                }) + utils.make_tag('a', {
-                    'class': 'line-number',
-                    'href': self.link_base + '#' + line_id,
-                }) + '</a>' + utils.make_tag('span', {
-                    'class': 'line-content',
-                }) + line.rstrip().replace('  ', '&nbsp; ') + '</span></span>\n')
+                yield 1, (utils.make_tag('span', {'class': 'line',
+                                                  'id': line_id})
+                          + (utils.make_tag('a', {
+                              'class': 'line-number',
+                              'href': self.link_base + '#' + line_id})
+                             + '</a>' if self.link_base else '')
+                          + utils.make_tag('span', {'class': 'line-content'})
+                          + line.rstrip('\n')
+                          + '</span></span>\n')
 
 
 class HtmlRenderer(misaka.HtmlRenderer):
-    """ Customized renderer for enhancing Markdown formatting
+    """ Customized renderer for enhanced Markdown formatting
 
-    Constructor arguments:
+    :param dict config: The configuration for the Markdown tags
+    :param list search_path: Directories to look in for resolving relatively-linked files
+    :param int entry_id: the numeric entry ID
+    :param list footnote_buffer: the buffer of footnote entries so far
+    :param list toc_buffer: the buffer of TOC entries so far
+    :param ItemCounter counter: item count accumulator
+
+    For configuration values, see the individual Markdown implementation functions,
+    :py:func:`HtmlRenderer.image`, and py:mod:`publ.image`.
 
-    config -- The configuration for the Markdown tags
-    search_path -- Directories to look in for resolving relatively-linked files
-    entry_id -- the numeric entry ID
-    footnote_buffer -- the buffer of footnote entries so far
-    toc_buffer -- the buffer of TOC entries so far
     """
 
     def __init__(self, args: typing.Dict,
@@ -142,9 +145,16 @@ class HtmlRenderer(misaka.HtmlRenderer):
 
         self._counter = counter
 
+    def _inner(self, text):
+        """ process some inner markdown """
+        return misaka.Markdown(self,
+                               self._config.get('markdown_extensions',
+                                                config.markdown_extensions))(text)
+
     @staticmethod
     def footnotes(_):
-        """ Actual footnote rendering is handled by the caller """
+        """ Stub function; footnote rendering is handled by
+        :py:func:entry.entry.footnotes """
         return None
 
     def _footnote_num(self, num):
@@ -161,7 +171,12 @@ class HtmlRenderer(misaka.HtmlRenderer):
                                     '#' + self._footnote_id(num, anchor))
 
     def footnote_ref(self, num):
-        """ Render a link to this footnote """
+        """ Render a link to a footnote
+
+        Relevant configuration options:
+        * ``footnotes_class``: The class to apply to a footnote marker
+        * ``footnotes_link``: The base URL for footnote links
+        """
         if self._config.get('_suppress_footnotes'):
             return '\u200b'  # zero-width space to prevent Misaka fallback
 
@@ -177,7 +192,14 @@ class HtmlRenderer(misaka.HtmlRenderer):
             content=self._footnote_num(num))
 
     def footnote_def(self, content, num):
-        """ Render the footnote body, deferring it if so configured """
+        """ Renders the body of a footnote to be used by :py:func:`entry.Entry.footnotes`.
+
+        Relevant configuration options:
+
+        * ``footnotes_link``: The base URL for footnote links
+        * ``footnotes_return``: The return symbol on an expanded footnote
+          (default: ``'↩'``)
+        """
         LOGGER.debug("footnote_def %d: %s", num, content)
 
         self._counter.footnote_def(content, num)
@@ -201,7 +223,7 @@ class HtmlRenderer(misaka.HtmlRenderer):
         self._footnote_buffer.append(text)
 
     def _header_id(self, content, level, num):
-        """ Return a reasonable anchor ID for a heading """
+        """ Return a reasonable anchor ID for a heading."""
         return '{eid}_h{level}_{num}_{slug}'.format(
             eid=self._entry_id,
             level=level,
@@ -209,7 +231,18 @@ class HtmlRenderer(misaka.HtmlRenderer):
             slug=slugify.slugify(content, max_length=32))
 
     def header(self, content, level):
-        """ Make a header with anchor """
+        """ Make a header with anchor.
+
+        Relevant configuration:
+
+        * ``toc_link``: The base URL for table of contents links from a heading
+        * ``heading_link_class``: The class to apply to a TOC link from a heading
+        * ``heading_link_attrs``: Additional attributes to apply to TOC links from headings
+        * ``heading_template``: The format template for a heading's TOC link
+          (default: ``{link}</a>{{text}}``). Receives the following template arguments:
+            * ``{link}``: The ``<a>`` tag linking to the TOC entry
+            * ``{text}``: The text of the heading
+         """
 
         self._counter.header(content, level)
         htag = f'h{level}'
@@ -222,7 +255,7 @@ class HtmlRenderer(misaka.HtmlRenderer):
             'href': urllib.parse.urljoin(self._config.get('toc_link', ''),
                                          '#' + hid),
             'class': self._config.get('heading_link_class', False),
-            **self._config.get('heading_link_config', {})
+            **self._config.get('heading_link_attrs', {})
         })
 
         if self._toc_buffer is not None:
@@ -245,27 +278,55 @@ class HtmlRenderer(misaka.HtmlRenderer):
     def image(self, raw_url, title='', alt=''):
         """ Adapt a standard Markdown image to a generated rendition set.
 
-        Container arguments used (in addition to the rendition tags):
+        Extended syntax::
+
+            ![alt text{container_args}](image1.jpg{img_args} "title text"
+            |image2.jpg{img_args} "title text"
+            |...)
+
+        Configuration for each image comes from the page template, then from the
+        ``container_args``, then from the ``img_args``.
+
+        Configuration that applies to the image set itself:
+
+        * ``figure``: If set, wrap the image set in a ``<figure>`` tag. If this
+          value is a string, use the string value as the CSS class name.
+
+        * ``caption``: If set, add the text as a ``<figcaption>``. Markdown supported.
+          Implies ``figure=True``.
+
+        * ``div_class``: The CSS class name to use on any wrapper div
+
+        * ``div_style``: Additional CSS styles to apply to the wrapper div
+
+        * ``count``: The maximum number of images to show at once
+
+        * ``more_text``: If there are more than ``count`` images, add this text
+          indicating that there are more images to be seen. This string gets two template
+          arguments, ``{count}`` which is the total number of images in the set, and
+          ``{remain}`` which is the number of images omitted from the set.
+
+        * ``more_link``: If `more_text` is shown, this will format the text as a link
+          to this location.
+
+        * ``more_class``: If `more_text` is shown, applies this CSS class to the link
+
+        See :py:mod:`publ.image` for configuration for image renditions.
 
-        div_class -- The CSS class name to use on any wrapper div
-        div_style -- Additional CSS styles to apply to the wrapper div
-        count -- The maximum number of images to show at once
-        more_text -- If there are more than `count` images, add this text indicating
-            that there are more images to be seen. This string gets two template
-            arguments, `{count}` which is the total number of images in the set,
-            and `{remain}` which is the number of images omitted from the set.
-        more_link -- If `more_text` is shown, this will format the text as a link to this location.
-        more_class -- If `more_text` is shown, wraps it in a `<div>` with this class.
         """
         # pylint: disable=too-many-locals
 
+        # If we aren't generating markup, there's no reason to do any of this
+        if not self._config.get('markup', True):
+            return ' '
+
         text = ''
 
         image_specs = raw_url
         if title:
             image_specs += f' "{title}"'
 
-        alt, container_args = image.parse_alt_text(alt)
+        alt, container_args = image.parse_img_config(alt)
 
         container_args = utils.prefix_normalize({**self._config, **container_args})
 
@@ -286,11 +347,10 @@ class HtmlRenderer(misaka.HtmlRenderer):
             if 'more_link' in container_args:
                 more_text = '{a}{text}</a>'.format(
                     text=more_text,
-                    a=utils.make_tag('a', {'href': container_args['more_link']}))
-            if 'more_class' in container_args:
-                more_text = '{div}{text}</div>'.format(
-                    text=more_text,
-                    div=utils.make_tag('div', {'class': container_args['more_class']}))
+                    a=utils.make_tag('a', {
+                        'href': container_args['more_link'],
+                        'class': container_args.get('more_class', False)
+                    }))
             text += flask.Markup(more_text)
 
         if text and (container_args.get('div_class') or
@@ -301,25 +361,60 @@ class HtmlRenderer(misaka.HtmlRenderer):
                                     'style': container_args.get('div_style') or False}),
                 text=text)
 
+        if text and (container_args.get('figure') or container_args.get('caption')):
+            fig_class = container_args.get('figure')
+            if container_args.get('caption'):
+                caption = '<figcaption>' + utils.strip_single_paragraph(
+                    self._inner(container_args['caption'])) + '</figcaption>'
+            else:
+                caption = ''
+            text = '{tag}{text}{caption}</figure>'.format(
+                tag=utils.make_tag('figure',
+                                   {'class': fig_class if isinstance(fig_class, str) else False}),
+                text=text,
+                caption=caption)
+
         # if text is ''/falsy then misaka interprets this as a failed parse...
         return text or ' '
 
     def blockcode(self, text, lang):
-        """ Pass a code fence through pygments """
+        """ Pass a code fence through pygments, and add line-numbering scaffolding.
+
+        Relevant configuration:
+
+        * ``code_highlight``: Whether to apply syntax highlighting to fenced
+          code blocks (default: ``True``)
+        * ``code_number_links``: Whether to add line-numbering links to fenced
+          code blocks' lines. If set to a string, specifies the base URL for
+          these links (default: ``True``)
+
+        """
         LOGGER.debug("blockcode lang=%s", lang)
 
         self._counter.blockcode(text, lang)
 
-        out = '\n<div class="blockcode">'
+        out = '<figure class="blockcode">'
+
+        lang, _, args = utils.parse_spec(lang, 0)
+        if args:
+            args = {**self._config, **args}
+        else:
+            args = self._config
+        LOGGER.debug("Code: language=%s args=%s", lang, args)
 
         if text.startswith('!'):
             caption, _, text = text.partition('\n')
-            caption = misaka.Markdown(TitleRenderer())(caption[1:].strip())
-            out += '<div class="caption">' + caption.strip() + '</div>'
+            caption = self._inner(caption[1:]).strip()
+            if caption:
+                out += '<figcaption>' + utils.strip_single_paragraph(caption) + '</figcaption>'
+        elif text.startswith(r'\!') or text.startswith('\n'):
+            # Allow the initial ! to be escaped out, or skip an initial blank line
+            text = text[1:]
 
-        out += '<pre>'
+        highlight = lang and args.get('code_highlight', True)
+        number_lines = highlight and args.get('code_number_links', True)
 
-        if lang and self._config.get('highlight_syntax', 'True'):
+        if highlight:
             try:
                 lexer = pygments.lexers.get_lexer_by_name(lang or 'text', stripall=True)
             except pygments.lexers.ClassNotFound:
@@ -327,24 +422,44 @@ class HtmlRenderer(misaka.HtmlRenderer):
         else:
             lexer = None
 
-        if lexer:
-            formatter = HtmlCodeFormatter(
-                line_id_prefix="e{}cb{}".format(self._entry_id, self._counter.code_blocks),
-                link_base=self._config.get('footnotes_link', ''),
-            )
-            out += '<code class="highlight">{}</code>'.format(
-                pygments.highlight(str(text), lexer, formatter))
-        else:
-            for line in text.split('\n'):
-                out += '<span class="line"><span class="line-content">{}</span></span>\n'.format(
-                    html.escape(line))
+        out += utils.make_tag('pre', {
+            'class': 'highlight' if highlight else False,
+            'data-language': lang if lang else False,
+            'data-line-numbers': None if number_lines else False,
+        } if lang else {})
 
-        out += '</pre></div>'
+        if lexer:
+            link_base = number_lines
+            if link_base is True:
+                # It was set explicitly True, which means we need to re-inherit the
+                # default from the template
+                link_base = self._config.get('code_number_links')
+            formatter = HtmlCodeFormatter(
+                line_id_prefix=f"e{self._entry_id}cb{self._counter.code_blocks}",
+                link_base=link_base,
+            )
+            out += pygments.highlight(str(text), lexer, formatter)
+        else:
+            lines = text.splitlines()
+            for line in lines:
+                out += ('<span class="line"><span class="line-content">'
+                        + html.escape(line) + '</span></span>\n')
+
+        out += '</pre></figure>'
 
         return out
 
     def link(self, content, link, title=''):
-        """ Emit a link, potentially remapped based on our embed or static rules """
+        """ Emit a link, potentially remapped based on our embed or static rules.
+
+        Relevant configuration:
+
+        * ``absolute``: Whether to produce absolute/external, rather than relative,
+          links (default: ``False``)
+        """
+
+        if not self._config.get('markup', True):
+            return content
 
         link = links.resolve(link, self._search_path,
                              self._config.get('absolute'))
@@ -358,16 +473,15 @@ class HtmlRenderer(misaka.HtmlRenderer):
 
     @staticmethod
     def paragraph(content):
-        """ emit a paragraph, stripping out any leading or following empty paragraphs """
+        """ emit a paragraph """
 
-        # if the content contains a top-level div then don't wrap it in a <p>
+        # if the content contains a top-level block element then don't wrap it in a <p>
         # tag
-        if content.startswith('<div') and content.endswith('</div>'):
-            return '\n' + content + '\n'
+        for element in ('div', 'figure'):
+            if content.startswith(f'<{element}') and content.endswith(f'</{element}>'):
+                return '\n' + content + '\n'
 
-        text = '<p>' + content + '</p>'
-        text = re.sub(r'<p>\s*</p>', r'', text)
-        return text or ' '
+        return '<p>' + content + '</p>'
 
     def _render_image(self, spec, show, container_args, alt_text=None):
         """ Render an image specification into an <img> tag """
@@ -418,7 +532,7 @@ def to_html(text, args, search_path,
     tocs: TocBuffer = toc_buffer if toc_buffer is not None else []
 
     # first process as Markdown
-    renderer = HtmlRenderer(args,
+    renderer = HtmlRenderer({**config.layout, **args},
                             search_path,
                             toc_buffer=tocs,
                             footnote_buffer=footnotes,
@@ -505,10 +619,10 @@ def render_title(text, markup=True, smartquotes=True, markdown_extensions=None):
     if smartquotes:
         text = misaka.smartypants(text)
 
-    if not markup:
-        text = html_entry.strip_html(text, remove_elements=PLAINTEXT_REMOVE_ELEMENTS)
+    if markup:
+        return flask.Markup(text)
 
-    return flask.Markup(text)
+    return html_entry.strip_html(text, remove_elements=PLAINTEXT_REMOVE_ELEMENTS)
 
 
 def toc_to_html(toc: TocBuffer, max_level: int = None) -> str:
