@@ -67,6 +67,8 @@ def handle_path_alias(path: str = None):
         return render_entry_record(alias.entry, alias.category, alias.template, _mounted=True)
     if isinstance(alias, path_alias.RenderCategory):
         return render_category_path(alias.category, alias.template)
+    if isinstance(alias, path_alias.AuthFailed):
+        return handle_unauthorized(alias.cur_user, category=alias.category)
 
     return None
 
@@ -111,7 +113,7 @@ def render_publ_template(template: Template, **kwargs) -> typing.Tuple[str, str]
         }
 
         text = template.render(**args)
-        return text, caching.get_etag(text), flask.g.stash
+        return text, caching.get_etag(text), flask.g.get('needs_auth')
 
     @orm.db_session
     def latest_entry():
@@ -125,10 +127,11 @@ def render_publ_template(template: Template, **kwargs) -> typing.Tuple[str, str]
         return None
 
     try:
-        flask.g.stash = {}
-        text, etag, flask.g.stash = do_render(
+        cur_user = user.get_active()
+        text, etag, flask.g.needs_auth = do_render(
             template,
-            user=user.get_active(),
+            user=cur_user,
+            _user_auth=cur_user.auth_groups if cur_user else None,
             _url=request.url,
             _index_time=index.last_indexed(),
             _latest=latest_entry(),
@@ -188,6 +191,23 @@ def render_exception(error):
 
     LOGGER.debug("render_exception %s %s", type(error), error)
 
+    if isinstance(error, http_error.Unauthorized):
+        from flask import current_app as app
+
+        force_ssl = config.auth.get('AUTH_FORCE_HTTPS')
+        if force_ssl and request.scheme != 'https':
+            return redirect(utils.secure_link(request.endpoint,
+                                              **request.view_args,
+                                              **request.args))
+
+        flask.g.needs_token = True
+        return app.authl.render_login_form(destination='/' + utils.redir_path(),
+                                           error=flask.g.get('token_error')), 401
+
+    result = handle_path_alias()
+    if result:
+        return result
+
     # Effectively strip off the leading '/', so map_template can decide
     # what the actual category is
     category = request.path[1:]
@@ -209,19 +229,6 @@ def render_exception(error):
                 'Refresh': retry
             })
 
-    if isinstance(error, http_error.Unauthorized):
-        from flask import current_app as app
-
-        force_ssl = config.auth.get('AUTH_FORCE_HTTPS')
-        if force_ssl and request.scheme != 'https':
-            return redirect(utils.secure_link(request.endpoint,
-                                              **request.view_args,
-                                              **request.args))
-
-        flask.g.needs_token = True
-        return app.authl.render_login_form(destination='/' + utils.redir_path(),
-                                           error=flask.g.get('token_error')), 401
-
     if isinstance(error, http_error.HTTPException):
         return render_error(category, error.name, error.code, exception={
             'type': type(error).__name__,
@@ -229,7 +236,7 @@ def render_exception(error):
             'args': error.args
         })
 
-    return render_error(category, "Exception occurred", 500, exception={
+    return render_error(category, "Exception Occurred", 500, exception={
         'type': type(error).__name__,
         'str': str(error),
         'args': error.args
@@ -272,13 +279,13 @@ def render_category(category='', template=None):
 
 def render_category_path(category: str, template: typing.Optional[str]):
     """ Renders the actual category by path """
+    import arrow
 
     if category:
         # See if there's any entries for the view...
-        if not orm.select(e for e in model.Entry  # type:ignore
-                          if (e.category == category
-                              or e.category.startswith(category + '/'))
-                          and e.visible):
+        if not model.Entry.select(lambda e: (e.category == category
+                                             or e.category.startswith(category + '/'))
+                                  and e.visible).exists():
             raise http_error.NotFound("No such category")
 
     if not template:
@@ -290,17 +297,23 @@ def render_category_path(category: str, template: typing.Optional[str]):
         # this might actually be a malformed category URL
         test_path = '/'.join((category, template)) if category else template
         LOGGER.debug("Checking for malformed category %s", test_path)
-        record = orm.select(
-            e for e in model.Entry if e.category == test_path).exists()  # type:ignore
+        record = model.Entry.select(lambda e: e.category ==
+                                    test_path and e.visible).exists()  # type:ignore
         if record:
-            return redirect(url_for('category', category=test_path, **request.args))
+            return redirect(url_for('category', category=test_path, **request.args),
+                            code=301)
 
         # nope, we just don't know what this is
         raise http_error.NotFound(f"No such view '{template}'")
 
     view_spec = view.parse_view_spec(request.args)
     view_spec['category'] = category
-    view_obj = view.View.load(view_spec)
+    try:
+        view_obj = view.View.load(view_spec)
+    except arrow.parser.ParserError as error:
+        raise http_error.BadRequest("Invalid date") from error
+    except queries.InvalidQueryError as err:
+        raise http_error.BadRequest(str(err))
 
     rendered, etag = render_publ_template(
         tmpl,
@@ -390,7 +403,7 @@ def _check_canon_entry_url(record):
             return result
 
         # Redirect to the canonical URL
-        return redirect(canon_url)
+        return redirect(canon_url, code=301)
 
     return None
 
@@ -409,7 +422,10 @@ def render_entry(entry_id, slug_text='', category=''):
     # pylint: disable=too-many-return-statements,too-many-branches,unused-argument
 
     # check if it's a valid entry
-    record = model.Entry.get(id=entry_id)
+    try:
+        record = model.Entry.get(id=entry_id)
+    except ValueError as error:
+        raise http_error.BadRequest("Invalid entry ID") from error
 
     # see if the file still exists
     if record and not os.path.isfile(record.file_path):
@@ -453,7 +469,7 @@ def render_entry_record(record: model.Entry, category: str, template: typing.Opt
     # if the entry canonically redirects, do that now
     if record.redirect_url:
         LOGGER.debug("Redirecting to %s", record.redirect_url)
-        return redirect(record.redirect_url)
+        return redirect(record.redirect_url, code=301)
 
     # check if the canonical URL matches
     if not _mounted:
@@ -500,8 +516,6 @@ def render_entry_record(record: model.Entry, category: str, template: typing.Opt
         'Content-Type': entry_obj.get('Content-Type', mime_type(tmpl)),
         'ETag': etag
     }
-    if record.status == model.PublishStatus.HIDDEN.value:
-        headers = {**headers, **NO_CACHE}
 
     return rendered, headers
 
@@ -515,11 +529,16 @@ def admin_dashboard(by=None):  # pylint:disable=invalid-name
 
     tmpl = map_template('', '_admin')
 
-    days = int(request.args.get('days', 0))
+    days = int(request.args.get('days', 7))
     count = int(request.args.get('count', 50))
     offset = int(request.args.get('offset', 0))
 
-    log, remain = user.auth_log(start=offset, count=count, days=days)
+    log, remain = user.auth_log(start=offset, count=count)
+
+    if 'user' in request.args:
+        focus_user = user.User(request.args['user'])
+    else:
+        focus_user = None
 
     rendered, _ = render_publ_template(
         tmpl,
@@ -529,7 +548,8 @@ def admin_dashboard(by=None):  # pylint:disable=invalid-name
         offset=offset,
         days=days,
         remain=remain,
-        by=by
+        by=by,
+        focus_user=focus_user,
     )
     return rendered
 
