@@ -1,3 +1,4 @@
+
 """ Authenticated user functionality """
 
 import ast
@@ -6,8 +7,10 @@ import configparser
 import datetime
 import logging
 import typing
+import urllib.parse
 
 import arrow
+import authl.disposition
 import flask
 from pony import orm
 from werkzeug.utils import cached_property
@@ -18,16 +21,16 @@ from .config import config
 LOGGER = logging.getLogger(__name__)
 
 
-@caching.cache.memoize(timeout=30)
-def get_groups(username: str, include_self: bool = True) -> typing.Set[str]:
-    """ Get the group membership for the given username """
+@caching.cache.memoize(timeout=5)
+def get_groups(identity: str, include_self: bool = True) -> typing.Set[str]:
+    """ Get the group membership for the given identity """
 
-    @caching.cache.memoize(timeout=30)
+    @caching.cache.memoize(timeout=10)
     def load_groups() -> typing.DefaultDict[str, typing.Set[str]]:
         # We only want empty keys; \000 is unlikely to turn up in a well-formed text file
         cfg = configparser.ConfigParser(delimiters=(
             '\000'), allow_no_value=True, interpolation=None)
-        # disable authentication lowercasing; usernames should only be case-densitized
+        # disable authentication lowercasing; identities should only be case-desensitized
         # by the auth backend
         cfg.optionxform = lambda option: option  # type:ignore
         cfg.read(config.user_list)
@@ -45,12 +48,12 @@ def get_groups(username: str, include_self: bool = True) -> typing.Set[str]:
     result: typing.Set[str] = set()
     pending: typing.Deque[str] = collections.deque()
 
-    pending.append(username)
+    pending.append(identity)
 
     while pending:
         check = pending.popleft()
         if check not in result:
-            if include_self or check != username:
+            if include_self or check != identity:
                 result.add(check)
             pending += groups.get(check, [])
 
@@ -60,30 +63,68 @@ def get_groups(username: str, include_self: bool = True) -> typing.Set[str]:
 class User(caching.Memoizable):
     """ An authenticated user """
 
-    def __init__(self, name: str,
+    def __init__(self, identity: str,
                  auth_type: typing.Optional[str] = None,
                  scope: typing.Optional[str] = None):
         """ Initialize the user object.
 
-        :param str name: The federated identity name
+        :param str identity: The federated identity name
         :param str auth_type: The authentication mechanism
         :param str scope: The user's access scope
         """
 
-        self._name = name
+        self._identity = identity
         self._auth_type = auth_type
         self._scope = scope
 
     def _key(self):
-        return self.name, self.auth_type, self.scope
+        return self._identity, self._auth_type, self._scope
 
     def __lt__(self, other):
-        return self.name < other.name
+        return self.identity < other.identity
 
     @cached_property
-    def name(self):
+    def identity(self):
         """ The federated identity name of the user """
-        return self._name
+        return self._identity
+
+    @cached_property
+    def humanize(self) -> str:
+        """ A humanized version of the identity string """
+        parsed = urllib.parse.urlparse(self._identity)
+        return ''.join(p for p in (
+            f'{parsed.scheme}:' if parsed.scheme not in ('http', 'https') else '',
+            parsed.netloc,
+            parsed.path,
+        ))
+
+    @cached_property
+    def name(self) -> str:
+        """ The readable name of the user """
+        if 'name' in self.profile:
+            return self.profile['name']
+
+        return self.humanize
+
+    @property
+    def profile(self) -> dict:
+        """ Get the user's profile """
+        return self._info[0]
+
+    @cached_property
+    def groups(self) -> typing.Set[str]:
+        """ The group memberships of the user, for display purposes """
+        return get_groups(self._identity, False)
+
+    @cached_property
+    def auth_groups(self) -> typing.Set[str]:
+        """ The group memberships of the user, for auth purposes """
+        return get_groups(self._identity, True)
+
+    @cached_property
+    def is_admin(self) -> bool:
+        """ Returns whether this user has administrator permissions """
+        return bool(config.admin_group and config.admin_group in self.groups)
 
     @cached_property
     def auth_type(self):
@@ -96,19 +137,36 @@ class User(caching.Memoizable):
         return self._scope
 
     @cached_property
-    def auth_groups(self) -> typing.Set[str]:
-        """ The group memberships of the user, for auth purposes """
-        return get_groups(self.name, True)
+    def _info(self) -> typing.Tuple[dict,
+                                    typing.Optional[datetime.datetime],
+                                    typing.Optional[datetime.datetime]]:
+        """ Gets the user info from the database
 
-    @cached_property
-    def groups(self) -> typing.Set[str]:
-        """ The group memberships of the user, for display purposes """
-        return get_groups(self.name, False)
+        :returns: profile, last_login, last_seen
+        """
+        with orm.db_session():
+            record = model.KnownUser.get(user=self._identity)
+            if record:
+                return (record.profile.copy(),
+                        record.last_login,
+                        record.last_seen)
+        return {}, None, None
 
     @property
-    def is_admin(self) -> bool:
-        """ Returns whether this user has administrator permissions """
-        return bool(config.admin_group and config.admin_group in self.groups)
+    def last_login(self) -> typing.Optional[arrow.Arrow]:
+        """ Get the latest known login time for the user, if any """
+        date = self._info[1]
+        return arrow.get(date).to(config.timezone) if date else None
+
+    @property
+    def last_seen(self) -> arrow.Arrow:
+        """ Get the latest known active time for the user """
+        date = self._info[2]
+        return arrow.get(date).to(config.timezone) if date else None
+
+    def token(self, lifetime: int, scope: str = None) -> str:
+        """ Get a bearer token for this user """
+        return tokens.get_token(self.identity, lifetime, scope)
 
 
 @utils.stash
@@ -135,21 +193,16 @@ def log_access(record, cur_user, authorized):
             'date': arrow.utcnow().datetime,
             'entry': record,
             'authorized': authorized,
-            'user': cur_user.name,
+            'user': cur_user.identity,
             'user_groups': str(cur_user.groups) if cur_user.groups else ''
         }
-        log_entry = model.AuthLog.get(entry=record, user=cur_user.name)
+        log_entry = model.AuthLog.get(entry=record, user=cur_user.identity)
         if log_entry:
             log_entry.set(**values)
         else:
             model.AuthLog(**values)
 
 
-@caching.cache.memoize(timeout=30)
-def _get_user(username):
-    return User(username)
-
-
 def _get_group_set(groups):
     try:
         return ast.literal_eval(groups)
@@ -157,20 +210,46 @@ def _get_group_set(groups):
         return set(groups.split(',')) if groups else set()
 
 
+@orm.db_session(retry=5)
+def register(verified: authl.disposition):
+    """ Registers a user from the on_verified Authl hook """
+    if isinstance(verified, authl.disposition.Verified):
+        LOGGER.info("Got login from user %s with profile %s", verified.identity, verified.profile)
+        identity = verified.identity
+        now = arrow.utcnow().datetime
+        values = {
+            'last_login': now,
+            'last_seen': now,
+            'profile': verified.profile
+        }
+
+        record = model.KnownUser.get(user=identity)
+        if record:
+            record.set(**values)
+        else:
+            record = model.KnownUser(user=identity, **values)
+
+        if (verified.profile
+            and 'endpoints' in verified.profile
+                and 'ticket_endpoint' in verified.profile['endpoints']):
+            tokens.send_auth_ticket(identity, flask.request.url_root,
+                                    verified.profile['endpoints']['ticket_endpoint'])
+
+
 @orm.db_session(retry=5)
 def log_user():
     """ Update the user table to see who's been by """
-    username = flask.session.get('me')
-    if username:
+    identity = flask.session.get('me')
+    if identity:
         values = {
             'last_seen': arrow.utcnow().datetime,
         }
 
-        record = model.KnownUser.get(user=username)
+        record = model.KnownUser.get(user=identity)
         if record:
             record.set(**values)
         else:
-            record = model.KnownUser(user=username, **values)
+            record = model.KnownUser(user=identity, **values)
 
 
 @orm.db_session()
@@ -181,7 +260,9 @@ def known_users(days=None):
         since = (arrow.utcnow() - datetime.timedelta(days=days)).datetime
         query = orm.select(e for e in query if e.last_seen >= since)
 
-    return [(_get_user(record.user), arrow.get(record.last_seen).to(config.timezone))
+    query = query.order_by(orm.desc(model.KnownUser.last_seen))
+
+    return [(User(record.user), arrow.get(record.last_seen).to(config.timezone))
             for record in query]
 
 
@@ -190,17 +271,13 @@ LogEntry = collections.namedtuple(
 
 
 @orm.db_session()
-def auth_log(days=None, start=0, count=100):
+def auth_log(start=0, count=100):
     """ Get the logged accesses to each entry """
-    query = model.AuthLog.select()
-    if days:
-        since = (arrow.utcnow() - datetime.timedelta(days=days)).datetime
-        query = orm.select(e for e in query if e.date >= since)
-    query = query.order_by(orm.desc(model.AuthLog.date))[start:]
+    query = model.AuthLog.select().order_by(orm.desc(model.AuthLog.date))[start:]
 
     return [LogEntry(date=arrow.get(record.date).to(config.timezone),
                      entry=record.entry,
-                     user=_get_user(record.user),
+                     user=User(record.user),
                      user_groups=_get_group_set(record.user_groups),
                      authorized=record.authorized)
             for record in query[:count]], len(query) - count
