@@ -1,6 +1,7 @@
 # utils.py
 """ Some useful utilities that don't belong anywhere else """
 
+import ast
 import functools
 import html
 import html.parser
@@ -12,8 +13,10 @@ import urllib.parse
 
 import arrow
 import flask
+import slugify
 import werkzeug.routing
 
+from . import model
 from .config import config
 
 LOGGER = logging.getLogger(__name__)
@@ -26,6 +29,8 @@ ListLike = typing.Union[typing.List[T],
 TagAttr = typing.Union[str, bool, None]
 TagAttrs = typing.Dict[str, TagAttr]
 
+_TagKey = slugify.Slugify(to_lower=True)
+
 
 class CallableProxy:
     """ Wrapper class to make args possible on properties. """
@@ -34,7 +39,7 @@ class CallableProxy:
     def __init__(self, func: typing.Callable[..., T]):
         """ Construct the property proxy.
 
-        func -- The function to wrap
+        :param func: The function to wrap
         """
 
         self._func: typing.Callable[..., T] = func
@@ -114,11 +119,9 @@ WEEK_FORMAT = 'YYYYMMDD_w'
 
 def parse_date(datestr: str) -> typing.Tuple[arrow.Arrow, str, str]:
     """ Parse a date expression into a tuple of:
+    ``(start_date, span_type, span_format)``
 
-        (start_date, span_type, span_format)
-
-    Arguments:
-        datestr -- A date specification, in the format of YYYY-MM-DD (dashes optional)
+    :param str datestr: A date specification, in the format of YYYY-MM-DD (dashes optional)
     """
 
     match = re.match(
@@ -147,10 +150,10 @@ def parse_date(datestr: str) -> typing.Tuple[arrow.Arrow, str, str]:
 def find_file(path: str, search_path: typing.Union[str, ListLike[str]]) -> typing.Optional[str]:
     """ Find a file by relative path. Arguments:
 
-    path -- the image's filename
-    search_path -- a list of directories to check in
+    :param str path: the image's filename
+    :param list search_path: a list of directories to check in
 
-    Returns: the resolved file path
+    :returns: the resolved file path
     """
 
     for relative in as_list(search_path):
@@ -166,8 +169,8 @@ def static_url(path: str, absolute: bool = False) -> str:
 
     Arguments:
 
-    path -- the path to the file (relative to the static files directory)
-    absolute -- whether the link should be absolute or relative
+    :param str path: the path to the file (relative to the static files directory)
+    :param bool absolute: whether the link should be absolute or relative
     """
 
     if os.sep != '/':
@@ -181,14 +184,12 @@ def make_tag(name: str,
              start_end: bool = False) -> str:
     """ Build an HTML tag from the given name and attributes.
 
-    Arguments:
-
-    name -- the name of the tag (p, div, etc.)
-    attrs -- a dict of attributes to apply to the tag
-    start_end -- whether this tag should be self-closing
+    :param str name: the name of the tag (p, div, etc.)
+    :param attrs: a dict or list of attributes to apply to the tag
+    :param bool start_end: whether this tag should be self-closing
 
     If an attribute's value is None it will be written as a standalone attribute,
-    e.g. <audio controls>. To suppress it entirely, make the value explicitly False.
+    e.g. ``<audio controls>``. To suppress it entirely, make the value explicitly False.
     """
 
     text = '<' + name
@@ -232,9 +233,12 @@ def file_fingerprint(fullpath: str) -> str:
 
 
 def remap_args(input_args: typing.Dict[str, typing.Any],
-               remap: typing.Dict[str, str]) -> typing.Dict[str, typing.Any]:
-    """ Generate a new argument list by remapping keys. The 'remap'
-    dict maps from destination key -> priority list of source keys
+               remap: typing.Dict[str, typing.Union[str, ListLike[str]]]
+               ) -> typing.Dict[str, typing.Any]:
+    """ Generate a new argument list by remapping keys.
+
+    :param dict input_args: The input argument list
+    :param dict remap: A mapping of destination key -> priority list of source keys
     """
     out_args = input_args
     for dest_key, src_keys in remap.items():
@@ -254,7 +258,14 @@ def remap_args(input_args: typing.Dict[str, typing.Any],
 
 
 def remap_link_target(path: str, absolute: bool = False) -> str:
-    """ remap a link target to a static URL if it's prefixed with @ """
+    """ Remap a link target from a path, URL, or static file reference.
+
+    :param str path: The original link target; start with ``@`` to indicate a
+        static file
+
+    :param bool absolute: Whether to make the link absolute with respect to the
+        current page URL
+    """
     if path.startswith('@'):
         # static resource
         return static_url(path[1:], absolute=absolute)
@@ -275,10 +286,11 @@ class HTMLTransform(html.parser.HTMLParser):
     """ Wrapper to HTMLParser to make it easier to build a SAX-style processor.
 
     You will probably want to implement:
-        handle_starttag(self, tag, attrs)
-        handle_endtag(self, tag)
-        handle_data(self, data)
-        handle_startendtag(self, tag, attrs)
+
+    * ``handle_starttag(self, tag, attrs)``
+    * ``handle_endtag(self, tag)``
+    * ``handle_data(self, data)``
+    * ``handle_startendtag(self, tag, attrs)``
     """
 
     def __init__(self):
@@ -290,9 +302,11 @@ class HTMLTransform(html.parser.HTMLParser):
         self._fed = []
 
     def feed(self, data: str):
-        """ Feed in some text data. Overrides the base class to ensure that
-        it's handled like a plain string and not a MarkupSafe string (which
-        causes double-escaping to happen) """
+        """ Feed in some text data.
+
+        Overrides the base class to ensure that it's handled like a plain string
+        and not a MarkupSafe string (which causes double-escaping to happen)
+        """
         super().feed(str(data))
 
     def append(self, item: str):
@@ -314,8 +328,26 @@ class HTMLTransform(html.parser.HTMLParser):
 
 
 def prefix_normalize(kwargs: ArgDict) -> ArgDict:
-    """ Given an argument list where one of them is 'prefix', normalize the
-    arguments to convert {prefix}{key} to {key} and remove the prefixed versions
+    """
+
+    Given an argument list where one of them is ``prefix``, normalize the
+    arguments to convert ``{prefix}{key}`` to ``{key}`` and remove the prefixed
+    versions
+
+    For example::
+
+        {
+            'prefix': 'index_',
+            'width': 320,
+            'index_width': 256,
+        }
+
+    becomes::
+
+        {
+            'width': 256
+        }
+
     """
 
     prefixed = {}
@@ -350,11 +382,13 @@ def as_list(item: typing.Any) -> ListLike:
 
 
 class CategoryConverter(werkzeug.routing.PathConverter):
-    """ A version of PathConverter that doesn't accept paths beginning with _ """
+    """ A version of PathConverter that only accepts valid Publ categories
+    (i.e. no segment starts with _) """
 
     def to_python(self, value: str) -> str:
-        if value[0] == '_':
-            raise werkzeug.routing.ValidationError
+        for part in value.split('/'):
+            if part[0] == '_':
+                raise werkzeug.routing.ValidationError
         return super().to_python(value)
 
 
@@ -370,8 +404,9 @@ class TemplateConverter(werkzeug.routing.UnicodeConverter):
 def redir_path(path: str = None) -> str:
     """ Convert a URI path to a path fragment, suitable for url_for
 
-    :param path: The path to redirect to; uses the current request.full_path if
-        unspecified
+    :param str path: The path to redirect to; uses the current request.full_path
+        if unspecified
+
     """
 
     if path is None:
@@ -388,8 +423,8 @@ def redir_path(path: str = None) -> str:
 
 
 def secure_link(endpoint: str, *args, **kwargs) -> str:
-    """ flask.url_for except it will force the link to be secure if we are
-    configured with AUTH_FORCE_HTTPS """
+    """ :py:func:`flask.url`_for except it will force the link to be secure if
+    we are configured with ``AUTH_FORCE_HTTPS`` """
     force_ssl = config.auth.get('AUTH_FORCE_HTTPS')
 
     if force_ssl and flask.request.scheme != 'https':
@@ -413,7 +448,7 @@ def auth_link(endpoint: str, auto_redir=True) -> typing.Callable[..., str]:
 
 
 def stash(func: typing.Optional[typing.Callable] = None):
-    """ Decorator to memoize a function onto the global context.
+    """ Decorator to memoize a function onto the request's global context.
     """
 
     def make_hashable(item):
@@ -423,9 +458,9 @@ def stash(func: typing.Optional[typing.Callable] = None):
         if isinstance(item, (list, tuple)):
             return tuple(make_hashable(i) for i in item)
         if isinstance(item, set):
-            return tuple(sorted(make_hashable(i) for i in item))
+            return frozenset(make_hashable(i) for i in item)
         if isinstance(item, dict):
-            return tuple(sorted((make_hashable(k), make_hashable(v)) for k, v in item.items()))
+            return frozenset((make_hashable(k), make_hashable(v)) for k, v in item.items())
         return item
 
     def decorator(inner: typing.Callable):
@@ -436,7 +471,7 @@ def stash(func: typing.Optional[typing.Callable] = None):
                 flask.g.store[inner] = {}
             store = flask.g.store[inner]
 
-            cache_key = make_hashable(args) + make_hashable(kwargs)
+            cache_key = (make_hashable(args), make_hashable(kwargs))
             if cache_key in store:
                 return store[cache_key]
 
@@ -458,16 +493,41 @@ def parse_tuple_string(argument: typing.Union[str, typing.Tuple, typing.List],
     return tuple(argument)
 
 
+@functools.lru_cache()
+def tag_key(tag):
+    """ Given a tag, return its normalized key """
+    if isinstance(tag, str):
+        return _TagKey(tag)
+    if isinstance(tag, model.EntryTag):
+        return tag.key
+    if isinstance(tag, model.EntryTagged):
+        return tag.tag.key
+    raise ValueError(f"TagSet got non-tag-type {type(tag)}")
+
+
+def tag_cname(tag) -> str:
+    """ Get the canonical name of a tag from a name or key """
+    record = model.EntryTag.get(key=tag_key(tag))
+    if record:
+        return record.name
+    return tag
+
+
 class TagSet(typing.Set[str]):
     """ A frozenset-equivalent class that is case-insensitive """
 
     def __init__(self, contents: ListLike[str] = None):
-        storage = {v.casefold(): v for v in contents} if contents else {}
-        self._keys = frozenset(storage.keys())
-        self._values = frozenset(storage.values())
+        super().__init__()
+        if contents:
+            storage = {tag_key(v): tag_cname(v) for v in contents}
+            self._keys = frozenset(storage.keys())
+            self._values = frozenset(storage.values())
+        else:
+            self._keys = frozenset()
+            self._values = frozenset()
 
     def __contains__(self, key) -> bool:
-        return key.casefold() in self._keys
+        return tag_key(key) in self._keys
 
     def __iter__(self):
         return self._values.__iter__()
@@ -486,19 +546,19 @@ class TagSet(typing.Set[str]):
 
     @staticmethod
     def _fold(items):
-        return {k.casefold() for k in items}
+        return {tag_key(k) for k in items}
 
     def __and__(self, other):
         folded = self._fold(other)
-        return TagSet((k for k in self if k.casefold() in folded))
+        return TagSet((k for k in self if tag_key(k) in folded))
 
     def __xor__(self, other):
         folded = self._fold(self) ^ self._fold(other)
-        return TagSet((k for k in self | other if k.casefold() in folded))
+        return TagSet((k for k in self | other if tag_key(k) in folded))
 
     def __sub__(self, other):
         folded = self._fold(other)
-        return TagSet((k for k in self if k.casefold() not in folded))
+        return TagSet((k for k in self if tag_key(k) not in folded))
 
     def __len__(self):
         return len(self._values)
@@ -519,3 +579,53 @@ class TagSet(typing.Set[str]):
 
     def __lt__(self, other):
         return self._fold(self) < self._fold(other)
+
+    def keys(self):
+        """ Return the case-folded names """
+        return self._keys
+
+    def values(self):
+        """ return the case-preserved names """
+        return self._values
+
+
+def strip_single_paragraph(text: str):
+    """ If the provided HTML text has only a single paragraph, strip it off. """
+
+    stripped = re.sub(r'^<p>(.*)</p>$', r'\1', text.strip())
+    if '<p>' in stripped:
+        return text
+    return stripped
+
+
+def parse_spec(text: str, pos_limit: int = None) -> typing.Tuple[str, list, ArgDict]:
+    """ Given a string like ``foo{10,bar=baz}``, parse out the argument lists.
+
+    :param str text: The text to parse
+    :param int pos_limit: The maximum number of positional args to allow
+    :returns: Tuple of text, pos_args, kw_args
+    """
+    match = re.match(r'([^\{]*)(\{(.*)\})$', text)
+    if match:
+        pos_args, kw_args = parse_arglist(match.group(3), pos_limit)
+        return match.group(1), pos_args, kw_args
+
+    return text, [], {}
+
+
+def parse_arglist(args: str, pos_limit: int = None) -> typing.Tuple[list, ArgDict]:
+    """ Parse an argument list into pos_args, kw_args"""
+    tree = ast.parse(f'f({args})')
+    expr = typing.cast(ast.Expr, tree.body[0])
+    funccall = typing.cast(ast.Call, expr.value)
+
+    pos_args = [ast.literal_eval(arg) for arg in funccall.args]
+    if pos_limit is not None and len(pos_args) > pos_limit:
+        raise TypeError(
+            f"Expected at most {pos_limit} positional args but {len(pos_args)} were given")
+
+    kwargs = {arg.arg: ast.literal_eval(arg.value)
+              for arg in funccall.keywords if arg.arg}
+
+    LOGGER.debug("pos_args=%s kw_args=%s", pos_args, kwargs)
+    return pos_args, kwargs
