@@ -121,23 +121,8 @@ class View(caching.Memoizable):
         """ Gets entries which are authorized for the current viewer """
 
         def _entries(unauthorized=0) -> typing.List[Entry]:
-            result: typing.List[Entry] = []
             count = self.spec.get('count')
-            cur_user = user.get_active()
-            for record in self._entries:
-                if count is not None and len(result) >= count:
-                    break
-
-                auth = record.is_authorized(cur_user)
-                if auth or unauthorized:
-                    result.append(Entry.load(record))
-                    if not auth and unauthorized is not True:
-                        unauthorized -= 1
-
-                if not auth:
-                    tokens.request(cur_user)
-
-            return result
+            return Entry.filter_auth(self._entries, count, unauthorized)
 
         return utils.CallableProxy(_entries)
 
@@ -163,6 +148,28 @@ class View(caching.Memoizable):
 
         return utils.CallableProxy(_unauthorized)
 
+    @cached_property
+    def has_unauthorized(self) -> bool:
+        """ Returns whether there's entries that the user could see if they
+        were authorized differently """
+
+        count = self.spec.get('count')
+        auth_count = 0
+        cur_user = user.get_active()
+        for record in self._entries:
+            if count is not None and auth_count >= count:
+                # we've already hit the end of the list
+                break
+
+            if not record.is_authorized(cur_user):
+                # there's at least one unauthorized, so fast quit
+                tokens.request(cur_user)
+                return True
+
+            auth_count += 1
+
+        return False
+
     @cached_property
     def deleted(self) -> typing.List[Entry]:
         """ Gets the deleted entries from the view """
@@ -332,9 +339,9 @@ class View(caching.Memoizable):
                             raise ValueError(f"key {k} is of type {type(val)}")
                         break
 
-            taglist = self.spec.get('tag')
+            taglist = kwargs.get('tag', self.spec.get('tag'))
             if taglist:
-                args['tag'] = taglist if isinstance(taglist, str) else list(taglist)
+                args['tag'] = list(utils.TagSet(utils.as_list(taglist)).keys())
 
             return flask.url_for('category',
                                  **args,
